package lieAlgebra;

import java.io.PrintStream;

public class CubicLieMarkovRateMatrix extends LieMarkovRateMatrix {
	
	/*
     * 
     * Licensed under the Creative Commons Zero (CC0) license (https://creativecommons.org/publicdomain/zero/1.0/)
	 * 
	 */
	private static final long serialVersionUID = 32806150698854471L;

	/*
	 * Given the model name, we can look up the basis matrices, and hence
	 * determine nParam_.
	 */
	public CubicLieMarkovRateMatrix(String modelName, String permutation) {
		super(modelName, permutation);
	}

	/**
	 * In CubicLieMarkovRateMatrix, the parameters (range (-1,1)) are used directly
	 * as weights of basis matrices (i.e. coordinates within a hypercube), and then 
	 * the whole is scaled to make the hypercube boundary coincide with the boundary 
	 * of stochasticity. 
	 */
	@Override
	public void createRelativeRates(double[][] rateStore,
			double[] rateParameters, int startIndex) {

		// Zero rateStore
		int dim = getDimension(); // 4
		for (int i=0; i<dim; i++) {
			for (int j=0; j<dim; j++) {
				rateStore[i][j] = 0;
			}
		}
		int n = matrices_.length;
		/* 'radius' r: r=0 is center of hyperpolyhedron (i.e. Jukes Cantor matrix A), r=1 is on
		 * the boundary of stochasticity, i.e. will have (at least) one off-diagonal equal to zero.
		 * As we will scale such that the boundary of the cube corresponds to the boundary of
		 * stochasticity, r = max abs value of parameters
		 */
		double r=0; 
		// Add weighted basis matrices. matrices_[0] is always matrix A, which we exclude at this point,
		// hence i starts at 1.
		for (int i=1; i<n; i++) {
			matrices_[i].applyToRates(rateStore, rateParameters[i+startIndex-1], permutation_);
			r = Math.max(r, Math.abs(rateParameters[i+startIndex-1]));
		}
		// find min off-diagonal (will always be negative, unless all off-diagonals are zero.)
		double min = Double.MAX_VALUE;
		for (int row=0; row<dim; row++) {
			for (int col=0; col<dim; col++) {
				if (col != row) min = Math.min(min,rateStore[row][col]);
			}
		}

		/*
		 *  If we call current matrix X', we want
		 *  X = k X' 
		 *  with k chosen such that minOffDiagonal(X) = -1,
		 *  then final rate matrix is
		 *  Q = (A+rX)/3
		 *  where r = radial parameter (first parameter) and A = Jukes Cantor rate matrix 
		 *  (and always-present basis matrix)
		 *  A = (-3  1  1  1)
		 *      ( 1 -3  1  1)
		 *      ( 1  1 -3  1)
		 *      ( 1  1  1 -3)
		 *     
		 */
		 // 'min' is always non-positive, r is always non-negative, hence scale is non-negative
		double scale = (min==0) ? 0 : -r/min/3; // min==0 iff r==0.
		for (int row=0; row<dim; row++) {
			for (int col=0; col<dim; col++) {				
				if (col == row) {
					rateStore[row][col]=rateStore[row][col]*scale-1;
				} else {
					rateStore[row][col]=rateStore[row][col]*scale+1./3;
				}
			}
		}		
	}

	/*
	 * Technically, the allowable range for all parameters is -1 to 1. However, 
	 * the boundary values give a rate matrix with zeros in it, which may lead to zero likelihood (hence
	 * infinite log likelihood) and ConjugateDirectionSearch in particular is both prone to
	 * evaluate on the boundary, and is unable to handle infinities. So I hack it a bit by
	 * setting the absolute maximum to a little less than one. 
	 */
	@Override
	public double getRateParameterLowerBound(int parameter) { return -0.99999; }
	@Override
	public double getRateParameterUpperBound(int parameter) { return  0.99999; }

	/*
	 *  Default parameters all zero => start at Jukes Cantor matrix
	 */
	@Override
	public void getDefaultRateParameters(double[] parameterStore, int startIndex) {
		int n = getNumberOfRateParameters();
		for (int i=startIndex; i<startIndex+n; i++) parameterStore[i] = 0;
	}

	// I've been lazy about superfluous ',' and end of row.
	@Override
	public void printObject(PrintStream out, String prefixString) {
		out.printf("%sCubicLieMarkovRateMatrix %s:\n",prefixString, name_);
		StringBuffer[][] matrix = new StringBuffer[4][4];
		for (int i=0; i<4; i++) for (int j=0; j<4; j++) matrix[i][j] = new StringBuffer();
		for (int i=0; i<=nParam_; i++) {
			matrices_[i].applyToStrings(matrix);
		}
		int maxlen=0;
		for (int i=0; i<4; i++) for (int j=0; j<4; j++) maxlen = Math.max(maxlen, matrix[i][j].length());
		String format = "%-"+Integer.toString(maxlen)+"s, ";
		for (int row=0; row<4; row++) {
			out.print(prefixString+'{');
			for (int col=0; col<4; col++) {
				out.printf(format,matrix[row][col].toString());
			}
			out.printf("}%n");
		}	
	}
}
