package lieAlgebra;

import java.io.PrintStream;
import java.util.HashMap;
import java.util.Map;

import pal.datatype.DataType;
import pal.datatype.Nucleotides;
import pal.substmodel.NeoRateMatrix;
/*
 * 
 * Licensed under the Creative Commons Zero (CC0) license (https://creativecommons.org/publicdomain/zero/1.0/)
 * 
 */

/*
 * Lie Markov models have rate matrices generated as a 
 * linear combination of basis matrices. As the basis matrices have
 * both positive and negative off-diagonal entries, there is a problem
 * in producing rate matrices which are guaranteed to be stochastic.
 * There are a number of parameterizations which solve this, which are
 * implemented as the subclasses of LieMarkovRateMatirx. 
 */

public abstract class LieMarkovRateMatrix implements NeoRateMatrix {
	private static final long serialVersionUID = -5096307261250540284L;

	protected String name_;
	protected int nParam_; 
	protected BasisMatrix[] matrices_; // nParam_+1 of them.
	// The multipliers for the basis matrices. Matches length of matrices_.
	// This is just preallocated storage space, need not be persistent.
	protected double[] basisMatrixWeights_; 
	protected Permutation permutation_; // RY, WS or MK
	/*
	 * Note that the 'rays' parameterization ignores matrices_ and uses basesMatrixWeights_
	 * and nParam_ somewhat differently. (Assuming I've got around to programming 'rays' at all...)
	 */
	
	
	/*
	 * Given the model name, we can look up the basis matrices, and hence
	 * determine nParam_.
	 */
	public LieMarkovRateMatrix(String modelName, String permutation) {
		name_ = modelName;
		permutation_ = Permutation.fromString(permutation);
		matrices_ = BasisMatrix.MODEL_BASIS_MATRICES.get(modelName);
		nParam_ = matrices_.length-1;
		basisMatrixWeights_ = new double[matrices_.length];
	}
	

	/*
	 * NeoRateMatrix methods
	 */
	/*
	 * For n dimensional model, takes (n-1) parameters.
	 * This always returns a rate matrix with trace = -4, removing one
	 * degree of freedom.
	 */
	public abstract void createRelativeRates(double[][] rateStore, double[] rateParameters, int startIndex);
	public abstract double getRateParameterLowerBound(int parameter);
	public abstract double getRateParameterUpperBound(int parameter);
	public abstract void getDefaultRateParameters(double[] parameterStore, int startIndex); 
	public abstract void printObject(PrintStream out, String prefixString);

	public String getUniqueName() { return name_; 	}
	public boolean isReversible() {	return false; 	}
	public boolean isIndependentOfEqbmFreq() { return false;	}
	public int getDimension() {		return 4;	}
	public int getNumberOfRateParameters() { return nParam_; }
	public boolean isDataTypeCompatible(DataType dt) { return dt.getNumStates()==4; }

	// Not yet coded output that processes parameters.
	public void printObject(PrintStream out, String prefixString, double[] parameters) {
		this.printObject(out, prefixString);
		out.printf("%s...unfinished,  sorry...%n",prefixString); 
	}

	public void printObject(PrintStream out) {
		this.printObject(out,"");
	}

	// TODO: enable these, and then make use of them in subclasses to simplify stuff.
	public void setUpperBounds(double[] bounds) {
		throw new RuntimeException("Can't alter parameter bounds");
	}


	public void setUpperBounds(double bound) {
		throw new RuntimeException("Can't alter parameter bounds");
	}
	
	
	//=====================================================================================
	/*
	 * Permutation inner class. Permutes rate matrix columns and rows
	 * to select which base pairs are the distinguished ones.
	 */
	
	// 'nA' for 'nucleotide A'. Avoids name collisions with basis matrices.
	private static final int nA = Nucleotides.A_STATE; 
	private static final int nC = Nucleotides.C_STATE;
	private static final int nG = Nucleotides.G_STATE;
	private static final int nT = Nucleotides.UT_STATE;
	
	private enum Permutation {
		RY (new int[]{nA,nG,nC,nT}),
		WS (new int[]{nA,nT,nC,nG}),
		MK (new int[]{nA,nC,nG,nT});
		
		protected final int[] permutation;
		Permutation(int[] perm) {
			this.permutation = perm;
		}
		public static Permutation fromString(String perm) {
			if (perm == null) return RY; // default to transitions vs transversions
			switch (perm) {
			case "RY":
				return RY;
			case "WS":
				return WS;
			case "MK":
				return MK;
			default:
				throw new IllegalArgumentException("Unrecognized permutation name "+perm);
			}
		}		
	}

	
	//=====================================================================================
	/*
	 * The basis matrices from which Lie Markov models are created. These will get used
	 * in createRelativeRates() method of subclasses of LieMarkovRateMatrix.
	 */

	protected static class BasisMatrix {
		private final int[][] positive; // indices of cells to add to. n x 2. Will add 1 to each cell unless isA2 is true.
		private final int[][] negative; // indices of cells to subtract from. Will subtract 1 from each cell unless isA is true
		private final boolean isA2; // for basis matrix A2, positive entries have weight 2
		private final boolean isA; // for basis matrix A, negative entries have weight 3
		public final String matrixName; // These two for display purposes only.
		public final String weightName;
		
		// can use this constructor for all except A and A2
		private BasisMatrix(int[][] pos, int[][] neg, String name) {
			this(pos,neg,false,false,name);
		}
		
		private BasisMatrix(int[][] pos, int[][] neg, boolean isA, boolean isA2, String name) {
			positive = pos;
			negative = neg;
			this.isA = isA;
			this.isA2 = isA2;
			matrixName = name;
			weightName = name.toLowerCase();
		}
		/*
		 * IMPORTANT NOTE!
		 * All the 'pictures' below are in columns-sum-to-zero order.
		 * pal works in rows-sum-to-zero, so these will get transposed in process of 
		 * building the rate matrix.
		 */
		
		protected void applyToRates(double[][] rateStore, double weight, Permutation perm) {
			// Note transposing coordinates in building rateStore, due to PAL's row/column convention 
			// differing from mine. (I.e. PAL uses row-sums-to-zero for rate matrix.)
			double wt = (isA) ? 3*weight : weight;
			for (int[] coord : negative) {
				rateStore[perm.permutation[coord[1]]][perm.permutation[coord[0]]] -= wt;
			}
			wt = (isA2) ? 2*weight : weight;
			for (int[] coord : positive) {
				rateStore[perm.permutation[coord[1]]][perm.permutation[coord[0]]] += wt;
			}
		}
		
		// applyToStrings is for display purposes. It is non-critical.
		protected void applyToStrings(StringBuffer[][] strings) {
			applyToStrings(strings, weightName);
		}
		protected void applyToStrings(StringBuffer[][] strings, String variable) {
			// Leave in column-sum order.
			String factor = isA2 ? "2" : "";
			for (int[] coord : positive) {
				strings[coord[0]][coord[1]].append('+').append(factor).append(variable);
			}
			factor = isA ? "3" : "";
			for (int[] coord : negative) {
				strings[coord[0]][coord[1]].append('-').append(factor).append(variable);
			}
		}
		
		/* 
		 * Basis matrix A (Jukes Cantor matrix) which is in all models.
		 * =+++
		 * +=++
		 * ++=+
		 * +++=
		 * where '=' means -3.
		 * 
		 * This basis matrix is hard-coded into createRelativeRates() method.
		 */
		
		public static final BasisMatrix A = new BasisMatrix(
				new int[][]{{0,1},{0,2},{0,3},{1,0},{1,0},{1,2},{1,3},{2,0},{2,1},{2,3},{3,0},{3,1},{3,2}},
				new int[][]{{0,0},{1,1},{2,2},{3,3}}, true, false, "A");

		/* 
		 * A1:
		 * -+00
		 * +-00
		 * 00-+
		 * 00+-
		 * 
		 * A1 and A2 are special cases. Only one or the other is ever used (A, A1 and A2 are
		 * not independent) and which is used depends on the parameterization chosen.
		 * A2 works best for the 'polar coordinate' parameterizations, A1 for the Fourier-Motzkin
		 * parameterization.  
		 */
		public static final BasisMatrix A1 = new BasisMatrix(
				new int[][]{{0,1},{1,0},{2,3},{3,2}},
				new int[][]{{0,0},{1,1},{2,2},{3,3}}, "A1");

		/*
		 * Basis matrix A2
		 * 0#--
		 * #0--
		 * --0#
		 * --#0
		 * where '#' means +2
		 * See comments on matrix A1 about the relationship between A1 and A2
		 */
		public static final BasisMatrix A2 = new BasisMatrix(
				new int[][]{{0,1},{1,0},{2,3},{3,2}},
				new int[][]{{0,2},{0,3},{1,2},{1,3},{2,0},{2,1},{3,0},{3,1}}, false, true, "A2");

		/*
		 * 00+-
		 * 00-+
		 * +-00
		 * -+00
		 */
		public static final BasisMatrix B = new BasisMatrix(
				new int[][]{{0,2},{1,3},{2,0},{3,1}},
				new int[][]{{0,3},{1,2},{2,1},{3,0}},"B");

		/*
		 * 00+-
		 * 00-+
		 * -+00
		 * +-00
		 */
		public static final BasisMatrix C = new BasisMatrix(
				new int[][]{{0,2},{1,3},{2,1},{3,0}},
				new int[][]{{0,3},{1,2},{2,0},{3,1}},"C");

		/*
		 * ++++
		 * ++++
		 * ----
		 * ----
		 */
		public static final BasisMatrix D = new BasisMatrix(
				new int[][]{{0,0},{0,1},{0,2},{0,3},{1,0},{1,1},{1,2},{1,3}},
				new int[][]{{2,0},{2,1},{2,2},{2,3},{3,0},{3,1},{3,2},{3,3}},"D");

		/*
		 * -+00
		 * +-00
		 * 00+-
		 * 00-+
		 */
		public static final BasisMatrix D1 = new BasisMatrix(
				new int[][]{{0,1},{1,0},{2,2},{3,3}},
				new int[][]{{0,0},{1,1},{2,3},{3,2}},"D1");

		/*
		 * E1 always pairs with E2
		 * ++++
		 * ----
		 * 0000
		 * 0000
		 */
		public static final BasisMatrix E1 = new BasisMatrix(
				new int[][]{{0,0},{0,1},{0,2},{0,3}},
				new int[][]{{1,0},{1,1},{1,2},{1,3}},"E1");
		/*
		 * E2 always pairs with E1
		 * 0000
		 * 0000
		 * ++++
		 * ----
		 */
		
		public static final BasisMatrix E2 = new BasisMatrix(
				new int[][]{{2,0},{2,1},{2,2},{2,3}},
				new int[][]{{3,0},{3,1},{3,2},{3,3}},"E2");

		/*
		 * F1 always pairs with F2
		 * ++--
		 * --++
		 * 0000
		 * 0000
		 */
		public static final BasisMatrix F1 = new BasisMatrix(
				new int[][]{{0,0},{0,1},{1,2},{1,3}},
				new int[][]{{1,0},{1,1},{0,2},{0,3}},"F1");
		/*
		 * F2 always pairs with F1
		 * 0000
		 * 0000
		 * ++--
		 * --++
		 */
		public static final BasisMatrix F2 = new BasisMatrix(
				new int[][]{{2,0},{2,1},{3,2},{3,3}},
				new int[][]{{3,0},{3,1},{2,2},{2,3}},"F2");

		/*
		 * G1 always pairs with G2
		 * +-00
		 * +-00
		 * -+00
		 * -+00
		 */
		public static final BasisMatrix G1 = new BasisMatrix(
				new int[][]{{0,0},{1,0},{2,1},{3,1}},
				new int[][]{{0,1},{1,1},{2,0},{3,0}},"G1");
		/*
		 * G2 always pairs with G1
		 * 00+-
		 * 00+-
		 * 00-+
		 * 00-+
		 */
		public static final BasisMatrix G2 = new BasisMatrix(
				new int[][]{{0,2},{1,2},{2,3},{3,3}},
				new int[][]{{0,3},{1,3},{2,2},{3,2}},"G2");
		/*
		 * And this is where the Lie Markov models are actually defined, in terms of their basis matrices.
		 * Can replace "A2" by "A1" throughout with mathematical legitimacy, and this will be done
		 * for the Fourier-Motzkin parameterizations.
		 */
		private static final Map<String,BasisMatrix[]> MODEL_BASIS_MATRICES = new HashMap<String,BasisMatrix[]>();
		static {
			MODEL_BASIS_MATRICES.put("1.1",   new BasisMatrix[]{A});
			MODEL_BASIS_MATRICES.put("2.2b",  new BasisMatrix[]{A,A2});
			MODEL_BASIS_MATRICES.put("3.3a",  new BasisMatrix[]{A,A2,B});
			MODEL_BASIS_MATRICES.put("3.3b",  new BasisMatrix[]{A,A2,C});
			MODEL_BASIS_MATRICES.put("3.3c",  new BasisMatrix[]{A,A2,D1});
			MODEL_BASIS_MATRICES.put("3.4",   new BasisMatrix[]{A,A2,D});
			MODEL_BASIS_MATRICES.put("4.4a",  new BasisMatrix[]{A,D, E1,E2});
			MODEL_BASIS_MATRICES.put("4.4b",  new BasisMatrix[]{A,A2,D, D1});
			MODEL_BASIS_MATRICES.put("4.5a",  new BasisMatrix[]{A,A2,B ,D});
			MODEL_BASIS_MATRICES.put("4.5b",  new BasisMatrix[]{A,A2,C ,D});
			MODEL_BASIS_MATRICES.put("5.6a",  new BasisMatrix[]{A,A2,B ,C ,D1});
			MODEL_BASIS_MATRICES.put("5.6b",  new BasisMatrix[]{A,A2,D ,E1,E2});
			MODEL_BASIS_MATRICES.put("5.7a",  new BasisMatrix[]{A,A2,B ,E1,E2});
			MODEL_BASIS_MATRICES.put("5.7b",  new BasisMatrix[]{A,A2,B ,F1,F2});
			MODEL_BASIS_MATRICES.put("5.7c",  new BasisMatrix[]{A,A2,B, G1,G2});
			MODEL_BASIS_MATRICES.put("5.11a", new BasisMatrix[]{A,A2,D1,E1,E2});
			MODEL_BASIS_MATRICES.put("5.11b", new BasisMatrix[]{A,A2,D1,F1,F2});
			MODEL_BASIS_MATRICES.put("5.11c", new BasisMatrix[]{A,A2,D1,G1,G2});
			MODEL_BASIS_MATRICES.put("5.16",  new BasisMatrix[]{A,A2,D ,G1,G2});
			MODEL_BASIS_MATRICES.put("6.6",   new BasisMatrix[]{A,A2,B, C, D, D1});
			MODEL_BASIS_MATRICES.put("6.7a",  new BasisMatrix[]{A,A2,B, D, E1,E2});
			MODEL_BASIS_MATRICES.put("6.7b",  new BasisMatrix[]{A,A2,C, D, E1,E2});
			MODEL_BASIS_MATRICES.put("6.8a",  new BasisMatrix[]{A,A2,D, D1,E1,E2});
			MODEL_BASIS_MATRICES.put("6.8b",  new BasisMatrix[]{A,A2,D, D1,G1,G2});
			MODEL_BASIS_MATRICES.put("6.17a", new BasisMatrix[]{A,A2,B, D, G1,G2});
			MODEL_BASIS_MATRICES.put("6.17b", new BasisMatrix[]{A,A2,C, D, G1,G2});
			MODEL_BASIS_MATRICES.put("8.8",   new BasisMatrix[]{A,A2,D, D1,E1,E2,F1,F2});
			MODEL_BASIS_MATRICES.put("8.10a", new BasisMatrix[]{A,A2,B, C, D, D1,E1,E2});
			MODEL_BASIS_MATRICES.put("8.10b", new BasisMatrix[]{A,A2,B, C, D, D1,G1,G2});
			MODEL_BASIS_MATRICES.put("8.16",  new BasisMatrix[]{A,A2,D, D1,E1,E2,G1,G2});
			MODEL_BASIS_MATRICES.put("8.17",  new BasisMatrix[]{A,A2,B, D, E1,E2,G1,G2});
			MODEL_BASIS_MATRICES.put("8.18",  new BasisMatrix[]{A,A2,B, D, E1,E2,F1,F2});
			MODEL_BASIS_MATRICES.put("9.20a", new BasisMatrix[]{A,A2,B, C, D1,E1,E2,F1,F2});
			MODEL_BASIS_MATRICES.put("9.20b", new BasisMatrix[]{A,A2,B, C, D1,F1,F2,G1,G2});
			MODEL_BASIS_MATRICES.put("10.12", new BasisMatrix[]{A,A2,B, C, D, D1,E1,E2,F1,F2});
			MODEL_BASIS_MATRICES.put("10.34", new BasisMatrix[]{A,A2,B, C, D, D1,E1,E2,G1,G2});
			MODEL_BASIS_MATRICES.put("12.12", new BasisMatrix[]{A,A2,B, C, D, D1,E1,E2,F1,F2,G1,G2});	
		}
	}
}
