package lieAlgebra;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.PushbackReader;
import java.util.Arrays;
import java.util.Calendar;

import pal.algorithmics.GeneralObjectState;
import pal.algorithmics.ProbabilityIterator;
import pal.algorithmics.Ranker;
import pal.algorithmics.SearchEngine;
import pal.algorithmics.StoppingCriteria;
import pal.algorithmics.UndoableAction;
import pal.alignment.AbstractAlignment;
import pal.alignment.Alignment;
import pal.alignment.AlignmentParseException;
import pal.alignment.ReadAlignment;
import pal.distance.DistanceMatrix;
import pal.distance.JukesCantorDistanceMatrix;
import pal.math.ConjugateDirectionSearch;
import pal.math.MersenneTwisterFast;
import pal.math.MultivariateMinimum;
import pal.math.OrthogonalSearch;
import pal.misc.Parameterized;
import pal.substmodel.LieMarkovModel;
import pal.substmodel.OrthogonalizedLieMarkovRateMatrix;
import pal.substmodel.RateMatrixHandler;
import pal.substmodel.RateMatrixSubstitutionModel;
import pal.substmodel.SubstitutionModel;
import pal.tree.NeighborJoiningTree;
import pal.tree.Tree;
import pal.tree.TreeTool;
import pal.tree.TreeUtils;
import pal.treesearch.PossiblyRootedMLSearcher;
import pal.util.AlgorithmCallback;
import Jama.Matrix;
/*
 * 
 * Licensed under the Creative Commons Zero (CC0) license (https://creativecommons.org/publicdomain/zero/1.0/)
 * 
 */

public class Main {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		
		// Parse command line.
		Configuration cfg = new Configuration(args);
		
		long startTime = Calendar.getInstance().getTimeInMillis();
		switch (cfg.main) {
		case 0: normalRun(cfg);
				break;
		case 1: RateMatrixHandler.test();
				break;
		case 2: LieMarkovModel.printOrbits();
			    OrthogonalizedLieMarkovRateMatrix.printModels();
				break;
		case 3: optTestRun(cfg);
				break;
		case 4: randomDebugStuff(cfg); // Whatever small bits of code I want to see run in the debugger.
				break;
		default: 	throw new IllegalArgumentException("Unrecognized -m option");
		}
		long endTime = Calendar.getInstance().getTimeInMillis();
		if (cfg.printTime) cfg.output.printf("Elapsed time %d ms\n", endTime-startTime);
		cfg.output.close();
	}
	
	
	
	private static void normalRun(Configuration cfg) {

		int nFiles = cfg.fileList.length;
		
		double[][] likelihoods = new double[nFiles][];
		for (int i=0; i<nFiles; i++) {
			String file = cfg.BASEDIR + cfg.fileList[i];
			try {
				likelihoods[i]=analyseFile(cfg, file, true);
			} catch (AlignmentParseException e) {
				cfg.output.printf("**** Failed to read and process file %s\n", file);
			}
			cfg.output.flush(); // shouldn't be necessary, having troubles with '-o' option not flushing stdout.
		}
		cfg.output.printf("%16s", cfg.printRelative ? "Relative logL" : "Absolute logL");
		for (String file: cfg.fileList) cfg.output.printf("%16s", file);
		cfg.output.println();
		
		String suffix = (cfg.gammaRateDistribution ? "+G" : "") + (cfg.invariableSites ? "+I" : "");
		for (int m=0; m<cfg.modelList.length; m++) {
			cfg.output.printf("%16s", cfg.modelList[m].name+suffix);
			for (int f=0; f<cfg.fileList.length; f++) {
				cfg.output.printf("%16.3f", 
						(m == 0 || !cfg.printRelative) ? likelihoods[f][m] : likelihoods[f][m] - likelihoods[f][0]);
			}
			cfg.output.println();
		}
	}
	
	// Run each model several times to test they always find the same optimum.
	private static void optTestRun(Configuration cfg) {
		int nTrials = cfg.nTrials;
		int nFiles = cfg.fileList.length;
		int nModels = cfg.modelList.length;
		MersenneTwisterFast rng = new MersenneTwisterFast(cfg.seed);
		double[][][] likelihoods = new double[nTrials][nFiles][];
		double base[] = new double[nFiles];
		for (int i=0; i<nFiles; i++) {base[i] = Double.NEGATIVE_INFINITY;}
		for (int trial=0; trial<nTrials; trial++) {
			// Randomize initial parameters for the models
			for (int model=0; model<nModels; model++) {
				randomizeTestModel(cfg.modelList[model],rng);
			}
			for (int file=0; file<nFiles; file++) {
				String filename = cfg.BASEDIR + cfg.fileList[file];
				try {
					likelihoods[trial][file]=analyseFile(cfg, filename, false);
				} catch (AlignmentParseException e) {
					cfg.output.printf("**** Failed to read and process file %s\n", file);
				}	
				for (int i=0; i<likelihoods[trial][file].length; i++) {
					base[file]=Math.max(base[file], likelihoods[trial][file][i]);
				}
			}
		}
		for (int model=0; model<nModels; model++) {
			cfg.output.printf("***** Results for model %s\n", cfg.modelList[model].name);
			for (int file=0; file<nFiles; file++) {
				cfg.output.printf("File %s, base likelihood %.4f\n", cfg.fileList[file], base[file]);
				for (int trial=0; trial<nTrials; trial++) {
					cfg.output.printf("%8.4f\n", base[file]-likelihoods[trial][file][model]);
				}
			}
		}
	}
	
	// Just do stuff that I want to see in the debugger. Edit at will.
	private static void randomDebugStuff(Configuration cfg) {
		RateMatrixSubstitutionModel model = Configuration.masterModelList.get(319).scModel;
		model.setParameter(0.499755659375,0);
		model.setParameter(0.5,1);
		model.setParameter(0.5,2);
		model.setParameter(0.5,3);
		double[][] q = new double[4][4];
		model.getRateMatrix(q);
		for (int i=0; i<4; i++) {
			System.out.println(java.util.Arrays.toString(q[i]));
		}
	}
		
	private static void randomizeTestModel(TestModel model, MersenneTwisterFast rng) {
		randomizeModel(model.scModel,rng);
		randomizeModel(model.gModel,rng);
		randomizeModel(model.iModel,rng);
		randomizeModel(model.giModel,rng);
	}
	
	private static void randomizeModel(RateMatrixSubstitutionModel model, MersenneTwisterFast rng) {
		int n=model.getNumParameters();
		for (int i=0; i<n; i++) {
			model.setParameter(rng.nextDouble(model.getLowerLimit(i), model.getUpperLimit(i)),i);
		}
	}
	
	
	
	private static double[] analyseFile(Configuration cfg, String filename, boolean resetModels) throws AlignmentParseException {
		cfg.output.printf("Analysing file %s\n",filename);
		cfg.output.flush(); // shouldn't be necessary, having troubles with '-o' option not flushing stdout.

		// Real data analysis. Step 1: read the data
		AbstractAlignment align=null;
		try {
			FileReader fr = new FileReader(filename);
			PushbackReader pbr = new PushbackReader(fr,20);
			align = new ReadAlignment(pbr);
			pbr.close();
			fr.close();
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		} catch (RuntimeException e) {
			e.printStackTrace();
			throw new AlignmentParseException("Alignment parsing failed with a RuntimeException");
		}
		cfg.output.printf("Alignment %d taxa, %d sites\n", align.getIdCount(), align.getSiteCount());
		cfg.output.flush(); // shouldn't be necessary, having troubles with '-o' option not flushing stdout.
		
		// Step 2: make an NJ tree
		DistanceMatrix dm = new JukesCantorDistanceMatrix(align);
		Tree njTree =  new NeighborJoiningTree(dm);
		// NJ can give negative branch lengths. Set these to zero
		TreeUtils.setMinBranchLength(njTree, 0);
		TreeUtils.forceBinary(njTree);
		njTree = TreeTool.getMidPointRooted(njTree);
		if (cfg.verbose){ 
			cfg.output.printf("NJ Tree:%s\n", njTree.toString());
			cfg.output.printf("NJ Tree diameter: %f\n",LMMTreeUtils.treeDiameter(njTree));
		}
		
		
		// Step 3: Optimise branch lengths and model parameters and output results
		int nModels = cfg.modelList.length;
		double[][] qArray = new double[4][4];
		Matrix qMatrix = new Matrix(qArray);
		double baseLikelihood = Double.NaN;
		double[] likelihoods = new double[nModels];
		for (int i=0; i<nModels; i++) {
			RateMatrixSubstitutionModel model;
			if      (!cfg.invariableSites && !cfg.gammaRateDistribution) {model = cfg.modelList[i].scModel;}
			else if ( cfg.invariableSites && !cfg.gammaRateDistribution) {model = cfg.modelList[i].iModel;}
			else if (!cfg.invariableSites &&  cfg.gammaRateDistribution) {model = cfg.modelList[i].gModel;}
			else    {                                                     model = cfg.modelList[i].giModel; }			
			String description = cfg.modelList[i].name + (cfg.gammaRateDistribution ? "+G" : "") + (cfg.invariableSites ? "+I" : "");
			Tree optTree = njTree.getCopy();

			switch (cfg.search) {
			case MIXED:     likelihoods[i] = maxLikelihoodMixed(optTree,align,model,description,cfg,resetModels);
							break;
			case FALLBACK:  likelihoods[i] = maxLikelihoodFallback(optTree,align,model,description,cfg,resetModels);
							break;
			case CONJUGATE: likelihoods[i] = maxLikelihoodConjugate(optTree,align,model,description,cfg,resetModels);
							break;
			case ORTHOGONAL:likelihoods[i] = maxLikelihoodOrthogonal(optTree,align,model,resetModels);
							break;
			case PARANOID:  likelihoods[i] = maxLikelihoodParanoid(optTree,align,model,description,cfg,resetModels);
							break;
			case APPROXIMATE:likelihoods[i]= maxLikelihoodOrthogonalApproximate(optTree,align,model,resetModels);
							break;				
			}
					
			double relativeLikelihood;
			if (Double.isNaN(baseLikelihood)) {
				baseLikelihood = likelihoods[i];
				relativeLikelihood = 0;
				cfg.output.printf("Base log likelihood=%f\n", baseLikelihood);
				cfg.output.flush(); // shouldn't be necessary, having troubles with '-o' option not flushing stdout.
			} else {
				relativeLikelihood = likelihoods[i]-baseLikelihood;
			}
			if (cfg.verbose) {
				cfg.output.printf("Tree:%s\n", optTree.toString());
				cfg.output.printf("Model %s (%d params) log likelihood = %f, rel log likelihood = %f\n",
						description,model.getNumParameters(),likelihoods[i],relativeLikelihood);
				cfg.output.printf("Tree diameter: %f\n",LMMTreeUtils.treeDiameter(optTree));
				cfg.output.print("MLE model parameters: ");
				double[] parameters = new double[model.getNumParameters()];
				for (int j=0; j<model.getNumParameters(); j++) {
					parameters[j]=model.getParameter(j);
					cfg.output.printf("%6.4f  ",parameters[j]);
				}
				cfg.output.println();	
				model.getRateMatrix(qArray);
				cfg.output.println("Q matrix:");
				qMatrix.print(cfg.output, 8, 4);
			}
		}
		return likelihoods;
	}
	
	/*
	 * 6 digits accuracy as default 
	 */
	public static double maxLikelihood(Tree tree, Alignment align, SubstitutionModel model, MultivariateMinimum optimiser) {
		return maxLikelihood(tree, align, model, optimiser, 6);
	}
	/*
	 * Largely cribbed from TreeSearchTool.optimiseUnrootedFixed
	 * A return value of 0 indicates a failure.
	 */
	public static double maxLikelihood(Tree tree, Alignment align, SubstitutionModel model, MultivariateMinimum optimiser, int nDigits) {
		// sanity check, values outside this range make little sense.
		if (nDigits<0 || nDigits>10) throw new IllegalArgumentException("Digits of accuracy should be between 0 and 10");
		// TODO: get seed from Configuration.
		long seed = 4;
		PossiblyRootedMLSearcher ut = new PossiblyRootedMLSearcher(tree, align, model, seed);
	    SearchEngine searcher = new SearchEngine( ProbabilityIterator.Utils.getHillClimb(), seed+1 );
	    final Ranker r = new Ranker(1);
        final UndoableAction action;
	    action = UndoableAction.Utils.getCombined(new UndoableAction[] {
				ut.getBranchLengthOptimiseAction( StoppingCriteria.Utils.getIterationCount( 2 ) ),
				ut.getModelOptimiseAction(optimiser, null, nDigits, nDigits ),
		} );
	    GeneralObjectState gos = new GeneralObjectState(action, ut, true );
	    double score;
	    // catch IllegalArgumentException: Non-finite entry in EigenSystem input matrix
	    try {
	    	searcher.run(
	    			AlgorithmCallback.Utils.getNullCallback(),  Double.NEGATIVE_INFINITY, gos,
	    			StoppingCriteria.Utils.getNonExactUnchangedScore( 3, false, 0.00001 ),
	    			r
	    	);
	    	score = searcher.getLastScore();
	    } catch (IllegalArgumentException e) {
	    	System.out.printf("Caught IllegalArgumentException: %s\n", e.getMessage());
			double[] errorParams = Parameterized.Utils.getParameters(model);
			System.out.printf("The parameters causing the error were %s\n", Arrays.toString(errorParams));
			score = 0;
	    }
	    // Create new tree from the altered branch lengths, and return it in 'tree':
	    tree.setRoot(ut.buildPALTree().getRoot());
        return score;
	}
	
	public static double maxLikelihoodOrthogonal(Tree tree, Alignment align, SubstitutionModel model, boolean resetModel) {
		if (resetModel) resetToDefault(model); // If we've analysed other files before this one, model will have been set to optimum for that data. 
		return maxLikelihood(tree, align, model, new OrthogonalSearch());
	}

	/*
	 * Low accuracy orthogonal search: two just two digits
	 */
	public static double maxLikelihoodOrthogonalApproximate(Tree tree, Alignment align, SubstitutionModel model, boolean resetModel) {
		if (resetModel) resetToDefault(model); // If we've analysed other files before this one, model will have been set to optimum for that data. 
		return maxLikelihood(tree, align, model, new OrthogonalSearch(), 2);
	}

	/**
	 * If it returns log likelihood == 0, the search failed.
	 * @param tree
	 * @param align
	 * @param model
	 * @param modelName
	 * @param cfg
	 * @return log likelihood
	 */
	public static double maxLikelihoodConjugate(
			Tree tree, 
			Alignment align, 
			SubstitutionModel model, 
			String modelName, 
			Configuration cfg,
			boolean resetModel
	) {
		if (resetModel) resetToDefault(model); // If we've analysed other files before this one, model will have been set to optimum for that data. 
		double likelihood = Double.NaN;
		// ConjugateDirectionSearch gets into an endless loop when called with only one 
		// dimension to optimizer over.
		if (model.getNumParameters()<=1) {
			likelihood = maxLikelihoodOrthogonal(tree, align, model, resetModel);
		} else {
			int failureCount = 10;
			MersenneTwisterFast rng = new MersenneTwisterFast(cfg.seed); // only used in case of singular matrix errors.
			do {
				try {
					likelihood = maxLikelihood(tree,align,model,new ConjugateDirectionSearch(cfg.seed));
					failureCount = 0;
				} catch (RuntimeException e) {
					if (e.getMessage().equals("Matrix is singular.")) {
						failureCount--;
						cfg.output.printf("Singular matrix error for model %s, will try up to %d more times.\n", 
								modelName, failureCount);
						int n = model.getNumParameters();
						for (int j=0; j<n; j++) model.setParameter(rng.nextDouble(),j);
						likelihood = Double.NaN; // in case this is the last try.
					} else {
						cfg.output.printf("Uncaught error for model %s\n",modelName);
						cfg.output.println(e.getMessage());
						cfg.output.println(e.getStackTrace().toString());
						likelihood=0;
						failureCount--;
					}
				}
			} while (failureCount > 0);
		}
		return likelihood; // ==0 iff failed all times
	}
	
	public static double maxLikelihoodFallback(
			Tree tree, 
			Alignment align, 
			RateMatrixSubstitutionModel model, 
			String modelName, 
			Configuration cfg, 
			boolean resetModel
	) {
		double likelihood = maxLikelihoodConjugate(tree, align, model, modelName, cfg, resetModel);
		if (likelihood == 0) {
			cfg.output.println("Conjugate direction search failed, falling back on orthogonal search");
			// if resetModels=false, should really reset it back to initial values.
			likelihood = maxLikelihoodOrthogonal(tree, align, model, resetModel);
			cfg.output.printf("Orthogonal direction found log likelihood = %f, will retry conjugate direction search from this point\n", likelihood);
			double newLikelihood = maxLikelihoodConjugate(tree, align, model, modelName, cfg, resetModel);
			if (newLikelihood==0) {
				cfg.output.println("Conjugate direction search failed again, so keeping orthogonal search value");
			} else {
				cfg.output.println("Conjugate direction search succeeded on retry, using its value");
				likelihood = newLikelihood;
			}
		}
		return likelihood;
	}
	
	public static double maxLikelihoodMixed(
			Tree tree, 
			Alignment align, 
			RateMatrixSubstitutionModel model, 
			String modelName, 
			Configuration cfg, 
			boolean resetModel
	) {
		double likelihood = maxLikelihoodOrthogonalApproximate(tree, align, model, resetModel);
		likelihood = maxLikelihoodConjugate(tree, align, model, modelName, cfg, false);
		return likelihood;
	}

	
	/**
	 * Similar to maxLikelihoodMixed, but alternates hill climb methods even without error indications
	 * @param tree
	 * @param align
	 * @param model
	 * @param modelName
	 * @param cfg
	 * @param resetModel
	 * @return
	 */
	public static double maxLikelihoodParanoid(
			Tree tree, 
			Alignment align, 
			RateMatrixSubstitutionModel model, 
			String modelName, 
			Configuration cfg, 
			boolean resetModel
	) {
		double[] initParams = Parameterized.Utils.getParameters(model);
		double likelihood = maxLikelihoodConjugate(tree, align, model, modelName, cfg, resetModel);
		if (likelihood == 0) {
			cfg.output.println("Conjugate direction search failed, get initial solution by orthogonal search");
			if (!resetModel) {
				// in this case, should start from initial provided model parameters, not from defaults
				for (int i=0; i<model.getNumParameters(); i++) model.setParameter(initParams[i], i);
			} // else maxLikelihoodOrthogonal will reset model to defaults 
			likelihood = maxLikelihoodOrthogonal(tree, align, model, resetModel);
			cfg.output.printf("First othrogonal search log likelihood =           %f\n",likelihood);
			likelihood = maxLikelihoodConjugate(tree, align, model, modelName, cfg, false);
			cfg.output.printf("First Conjugate direction search log likelihood =  %f\n",likelihood);
			likelihood = maxLikelihoodOrthogonal(tree, align, model, false);
			cfg.output.printf("Second orthogonal search log likelihood =          %f\n",likelihood);
		} else {
			cfg.output.printf("First Conjugate direction search log likelihood =  %f\n",likelihood);
			likelihood = maxLikelihoodOrthogonal(tree, align, model, false);
			cfg.output.printf("First orthogonal search log likelihood =           %f\n",likelihood);
			likelihood = maxLikelihoodConjugate(tree, align, model, modelName, cfg, false);
			cfg.output.printf("Second Conjugate direction search log likelihood = %f\n",likelihood);
		}
		return likelihood;
	}

	
	/*
	 * Really belongs somewhere else. I've made a copy in palExtensions project,
	 * but don't want to link that or I end up with two different PAL libraries.
	 */
	public static void resetToDefault(SubstitutionModel model) {
		int n=model.getNumParameters();
		for (int i=0; i<n; i++) {
			model.setParameter(model.getDefaultValue(i),i);
		}
	}
	
	/*
	 * Optimize branch lengths only. For debugging.
	 */
	public static double maxLikelihoodBranches(Tree tree, Alignment align, SubstitutionModel model) {
		// TODO: get seed from Configuration.
		long seed = 4;
		PossiblyRootedMLSearcher ut = new PossiblyRootedMLSearcher(tree, align, model, seed);
	    SearchEngine searcher = new SearchEngine( ProbabilityIterator.Utils.getHillClimb(), seed+1 );
	    final Ranker r = new Ranker(1);
        final UndoableAction action;
	    action = ut.getBranchLengthOptimiseAction( StoppingCriteria.Utils.getIterationCount( 2 ) );
	    GeneralObjectState gos = new GeneralObjectState(action, ut, true );
        searcher.run(
        	AlgorithmCallback.Utils.getNullCallback(),  Double.NEGATIVE_INFINITY, gos,
            StoppingCriteria.Utils.getNonExactUnchangedScore( 3, false, 0.00001 ),
			r
        );
        return searcher.getLastScore();
	}
}
