package lieAlgebra;

import java.io.PrintStream;
/*
 * 
 * Licensed under the Creative Commons Zero (CC0) license (https://creativecommons.org/publicdomain/zero/1.0/)
 * 
 */


public class OctahedralLieMarkovRateMatrix extends LieMarkovRateMatrix {
	private static final long serialVersionUID = -4843344396374304999L;

	/*
	 * Given the model name, we can look up the basis matrices, and hence
	 * determine nParam_.
	 */
	public OctahedralLieMarkovRateMatrix(String modelName, String permutation) {
		super(modelName, permutation);
	}
	
	@Override
	public void createRelativeRates(double[][] rateStore,
			double[] rateParameters, int startIndex) {
		// set basis matrix weights. First parameter is radial coordinate, so skip that.
		// First basis matrix is A, which we also skip
		OctahedralVector.parametersToCoordinates(nParam_, rateParameters, startIndex+1, basisMatrixWeights_, 1);
		// Zero rateStore
		int dim = getDimension(); // 4
		for (int i=0; i<dim; i++) {
			for (int j=0; j<dim; j++) {
				rateStore[i][j] = 0;
			}
		}
		int n = basisMatrixWeights_.length;
		// Add weighted basis matrices (not including matrix A, hence start at i=1)
		for (int i=1; i<n; i++) {
			matrices_[i].applyToRates(rateStore, basisMatrixWeights_[i], permutation_);
		}
		// find min off-diagonal (will always be negative, unless all off-diagonals are zero.)
		double min = Double.MAX_VALUE;
		for (int row=0; row<4; row++) {
			for (int col=0; col<4; col++) {
				if (col != row) min = Math.min(min,rateStore[row][col]);
			}
		}
		/*
		 *  If we call current matrix X', we want
		 *  X = k X' 
		 *  with k chosen such that minOffDiagonal(X) = -1,
		 *  then final rate matrix is
		 *  Q = (A+rX)/3
		 *  where r = radial parameter (first parameter) and A = Jukes Cantor rate matrix 
		 *  (and always-present basis matrix)
		 *  A = (-3  1  1  1)
		 *      ( 1 -3  1  1)
		 *      ( 1  1 -3  1)
		 *      ( 1  1  1 -3)
		 *     
		 */
		// 'min' is always non-positive, rP[sI] is always non-negative, hence scale is non-negative
		double scale = (min==0) ? 0 : -rateParameters[startIndex]/min/3; // avoid NaN if min==0.
		for (int row=0; row<4; row++) {
			for (int col=0; col<4; col++) {				
				if (col == row) {
					rateStore[row][col]=rateStore[row][col]*scale-1;
				} else {
					rateStore[row][col]=rateStore[row][col]*scale+1./3;
				}
			}
		}		
	}
	
	/*
	 * Technically, the allowable range for all parameters is -1 to 1. However, values on the
	 * boundaries lead to zeros in the rate matrix, which often leads to zero likelihood (hence
	 * infinite log likelihood) and ConjugateDirectionSearch in particular is both prone to
	 * evaluate on the boundary, and is unable to handle infinities. So I hack it a bit by
	 * setting the range a little smaller than -1 to 1.
	 * 
	 * Parameter 0 is the 'radius': how far from Jukes Cantor rate matrix we are. If this equals zero,
	 * all other parameters are irrelevant. I give it a non-zero minimum so that the other
	 * parameters will always have some effect, else we might just get stuck on parameter[0]=0
	 * and never be able to move off it as we never try a favourable direction.
	 * 
	 */
	@Override
	public double getRateParameterLowerBound(int parameter) { return (parameter==0) ? 1e-5 : -0.99999; }
	@Override
	public double getRateParameterUpperBound(int parameter) { return  0.99999; }
	// Choice of default parameters is pretty arbitrary
	@Override
	public void getDefaultRateParameters(double[] parameterStore, int startIndex) {
		int n = getNumberOfRateParameters();
		for (int i=startIndex; i<startIndex+n; i++) parameterStore[i] = 0.5;
	}

	// I've been lazy about superfluous ',' and end of row.
	@Override
	public void printObject(PrintStream out, String prefixString) {
		out.printf("%sOctahedralLieMarkovRateMatrix %s:\n",prefixString, name_);
		StringBuffer[][] matrix = new StringBuffer[4][4];
		for (int i=0; i<4; i++) for (int j=0; j<4; j++) matrix[i][j] = new StringBuffer();
		for (int i=0; i<=nParam_; i++) {
			matrices_[i].applyToStrings(matrix);
		}
		int maxlen=0;
		for (int i=0; i<4; i++) for (int j=0; j<4; j++) maxlen = Math.max(maxlen, matrix[i][j].length());
		String format = "%-"+Integer.toString(maxlen)+"s, ";
		for (int row=0; row<4; row++) {
			out.print(prefixString+'{');
			for (int col=0; col<4; col++) {
				out.printf(format,matrix[row][col].toString());
			}
			out.printf("}%n");
		}	
	}

	public static void printModels() {
		new OctahedralLieMarkovRateMatrix("1.1",  "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("2.2b", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("3.3a", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("3.3b", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("3.3c", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("3.4",  "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("4.4a", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("4.4b", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("4.5a", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("4.5b", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("5.6a", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("5.6b", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("5.7a", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("5.7b", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("5.7c", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("5.11a","RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("5.11b","RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("5.11c","RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("5.16", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("6.6",  "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("6.7a", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("6.7b", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("6.8a", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("6.8b", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("6.17a","RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("6.17b","RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("8.8",  "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("8.10a","RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("8.10b","RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("8.16", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("8.17", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("8.18", "RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("9.20a","RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("9.20b","RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("10.12","RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("10.34","RY").printObject(System.out);
		new OctahedralLieMarkovRateMatrix("12.12","RY").printObject(System.out);
	}
}
