package lieAlgebra;

import pal.misc.Parameterized;
/*
 * 
 * Licensed under the Creative Commons Zero (CC0) license (https://creativecommons.org/publicdomain/zero/1.0/)
 * 
 */

/**
 * An octahedral vector is one where the sum of the absolute values of the entries equals 1.
 * 
 * This class provides a function to map (n-1) arguments (x_i) in the range [-1,1] to a length n
 * octahedral vector (y_j), with the following properties:
 * * The mapping is continuous and monotonic (except if one parameter = +/- 1, all subsequent parameters are irrelevant) 
 * * All possible arguments map to an octahedral vector
 * * All possible octahedral vectors are within the domain of the mapping.
 * 
 * Parameter x_0 simply specifies coordinate y_0.
 * Given coordinates (y_0,...,y_{i-1}) we have residual r_i = 1 - sum(abs(y_j),j=0..i-1)
 * and then y_i = x_i r_i.
 * This works until the last parameter x_{n-2}, which must specify both y_{n-2} and y_{n-1}.
 * y_{n-2} = (2 abs(x_{n-2})-1) r_{n-2}
 * y_{n-1} = sign(y_{n-2}) r_{n-1}
 * 
 * Note that we could got a similar result with polar coordinates (less the radial coordinate)
 * Either way, we are using (n-1) parameters to specify a direction in n dimensional space 
 * (equivalently, a point on the surface of an n-sphere or n-cross-polytope centered on the origin).
 * The octahedral coordinates are computationally cheaper. 
 * 
 * The inverse of the mappings is provided.
 * 
 * Can be used as a purely static class, providing transformations in both directions.
 * Can also be an object, where we can set parameters and get coordinates, or vice-versa.
 * 
 * TODO: Tests
 * 
 * @author woodhams
 *
 */

public class OctahedralVector implements Parameterized {
	double[] coordinates_; // 'y' in the comments above
	double[] parameters_; // 'x' in the comments above
	int n_; // = vector_.length;
	
	/**
	 * @param dimensions: a vector with non-negative elements summing to one
	 * 
	 * NOTE: dimensions array is call-by-reference. The passed array can
	 * have its contents changed by this object. 
	 */
	public OctahedralVector(double[] coordinates) {
		allocateSpace(coordinates.length);
		setCoordinates(coordinates);
		coordinatesToParameters(coordinates_, parameters_, 0);
	}
	
	public OctahedralVector clone() {
		return new OctahedralVector(this.coordinates_);
	}
	
	/**
	 * @return The octahedral vector
	 */
	public double[] getCoordinates() {
		return coordinates_;
	}
	/**
	 * copies the octahedral vector into the store.
	 */
	public void getCoordinates(double[] store) {
		parametersToCoordinates(parameters_,store);
	}
	/**
	 * Set the octahedral vector (validity of vector not checked.)
	 * @param coordinates
	 * 
	 * Note: Unlike the constructor, this copies the input (call by value).
	 * Is there some logical reason for this, or is it an inadvertent inconsistency?
	 */
	public void setCoordinates(double[] coordinates) {
		allocateSpace(coordinates.length);
		coordinates_ = coordinates.clone();
		coordinatesToParameters(coordinates_, parameters_, 0);
	}

	/**
	 * @return The length of the stochastic vector (one more than the number of parameters)
	 */
	// c.f. getNumParameters()
	public int length() {
		return n_;
	}
	/**
	 * If vector length has changed, updates n_ and allocates a new parameters_ array.
	 * Does NOT change coordinates_ array, as whatever is calling allocateSpace takes care of this.
	 * @param newLen
	 */
	private void allocateSpace(int newLen) {
		if (newLen != n_) {
			n_ = newLen;
			parameters_ = new double[n_-1];
		}
	}
	
	/**
	 * Methods for Parameterized interface:
	 */
	public double getDefaultValue(int n) { return 0.5; } // A pretty arbitrary choice. There is no 'neutral' parameter setting.
	public double getLowerLimit(int n)   { return -1; }
	public double getUpperLimit(int n)   { return 1; }
	public int getNumParameters()        { return n_-1; }
	public double getParameter(int n)    { return parameters_[n]; }
	public void setParameter(double param, int n) {
		parameters_[n] = param;
		parametersToCoordinates(parameters_,coordinates_);
	}
	public void setParameterSE(double paramSE, int n) { throw new RuntimeException("Not implemented");	}
	
	
	/* **************************************************************************************************************
	 * The static methods:
	 */
	
	/**
	 * The most general method. 
	 * @param n The length of the octahedral vector to be generated
	 * @param parameters A vector containing within it the n-1 parameters to be mapped to the octahedral vector 
	 * @param parametersOffset The starting point within the parameters vector 
	 * @param coordinates A vector to store the results within
	 * @param coordinatesOffset Where within vector to start storing results.
	 */
	public static void parametersToCoordinates(int n, double[] parameters, int parametersOffset, double[] coordinates, int coordinatesOffset) {
		double residual = 1;
		double value;
		for (int i=0; i<n-2; i++) {
			value = residual * parameters[i+parametersOffset];
			coordinates[i+coordinatesOffset] = value;
			residual -= Math.abs(value);
		}
		value = (2*Math.abs(parameters[parametersOffset+n-2])-1)*residual;
		coordinates[coordinatesOffset+n-2] = value;
		residual -= Math.abs(value);		
		coordinates[coordinatesOffset+n-1]  = Math.signum(parameters[parametersOffset+n-2]) * residual;
	}
	
	/**
	 * Write octahedral vector into the start of provided storage space,
	 * using all of the provided parameters vector. 
	 * @param parameters
	 * @param coordinates
	 */
	public static void parametersToCoordinates(double[] parameters, double[] coordinates) {
		parametersToCoordinates(parameters.length+1, parameters, 0, coordinates, 0);
	}
	
	public static double[] parametersToCoordinates(double[] parameters) {
		double[] coordinates = new double[parameters.length+1];
		parametersToCoordinates(parameters, coordinates);
		return coordinates;
	}
	
	/**
	 * Inverse mapping: given an octahedral vector, find the corresponding parameters.
	 * @param n The length of the octahedral vector supplied
	 * @param coordinates A vector containing within it the octahedral vector 
	 * @param coordinatesOffset Where to find the octahedral vector within 'coordinates'
	 * @param parameters A vector to store the results within
	 * @param parametersOffset Where within parameters array to start storing results. (Will store (n-1) numbers.)
	 * 
	 * If coordinates to not have absolute values summing to one, will still produce a result
	 * (without throwing an exception) which will, on reverse transformation, match the input
	 * for as many initial entries as is possible.
	 */
	public static void coordinatesToParameters(int n, double[] coordinates, int coordinatesOffset, double[] parameters, int parametersOffset) {
		double cumulativeSum = 0; // cumulative sum of absolute values
		// Extra complexity ('if' statements) ensures all values strictly in [-1,1] range even
		// in the presence of rounding errors or a sequence of zeros at the end of 'coordinates'.
		for (int i=0; i<n-2; i++) {		
			if (cumulativeSum <1) {
				double value = coordinates[i+coordinatesOffset];
				parameters[i+parametersOffset] = value/(1-cumulativeSum);
				cumulativeSum += Math.abs(value);
				if (cumulativeSum > 1) {
					parameters[i+parametersOffset] = Math.signum(value);
				}
			} else {
				// any value would do here 
				parameters[i+parametersOffset] = 0;
			}
		}
		if (cumulativeSum<1) {
			double value = coordinates[coordinatesOffset+n-2];
			double rawParam = value/(1-cumulativeSum);
			if (Math.abs(rawParam)>1) rawParam = Math.signum(rawParam);
			// Transform from range [-1,1] to [0,1] and set sign according to last coordinate
			// (can't use 'Math.signum' as it will cause bug if last coordinate == 0)
			if (coordinates[coordinatesOffset+n-1]<0) {
				parameters[parametersOffset+n-2] = (rawParam+1)/2;
			} else {
				parameters[parametersOffset+n-2] = -(rawParam+1)/2;
			}
		} else {
			// any values would do here
			parameters[parametersOffset+n-2]=0;
		}
	}
	/**
	 * Inverse mapping: given an octahedral vector, find the corresponding parameters.
	 * @param coordinates The octahedral vector (of length n)
	 * @param parameters A vector to store the parameters within 
	 * @param parametersOffset Where within vector to start storing results. (Will store (n-1) numbers.)
	 */
	public static void coordinatesToParameters(double[] coordinates, double[] parameters, int parametersOffset) {
			coordinatesToParameters(coordinates.length, coordinates, 0, parameters, parametersOffset);
	}
}
