package lieAlgebra;

import java.io.PrintStream;
/*
 * 
 * Licensed under the Creative Commons Zero (CC0) license (https://creativecommons.org/publicdomain/zero/1.0/)
 * 
 */

/*
 * Very similar to OctahedralLieMarkovRateMatrix, except uses SphericalVector in place
 * of OctahedralVector.
 */

public class SphericalLieMarkovRateMatrix extends LieMarkovRateMatrix {
	private static final long serialVersionUID = 5349001517171726419L;

	/*
	 * Given the model name, we can look up the basis matrices, and hence
	 * determine nParam_.
	 */
	public SphericalLieMarkovRateMatrix(String modelName, String permutation) {
		super(modelName, permutation);
	}
	
	@Override
	public void createRelativeRates(double[][] rateStore,
			double[] rateParameters, int startIndex) {
		// set basis matrix weights. First parameter is radial coordinate, so skip that.
		// First basis matrix is A, which we also skip
		SphericalVector.parametersToCoordinates(nParam_, rateParameters, startIndex+1, basisMatrixWeights_, 1);
		// Zero rateStore
		int dim = getDimension(); // 4
		for (int i=0; i<dim; i++) {
			for (int j=0; j<dim; j++) {
				rateStore[i][j] = 0;
			}
		}
		int n = basisMatrixWeights_.length;
		// Add weighted basis matrices (not including matrix A, hence start at i=1)
		for (int i=1; i<n; i++) {
			matrices_[i].applyToRates(rateStore, basisMatrixWeights_[i], permutation_);
		}
		/*
		 * For SphericalLieMarkovMatrix, the radial parameter (rateParameters[startIndex])
		 * will always be non-negative, and this 'if' will have no effect. It is
		 * included for the sake of subclass ExtendedSphericalLieMarkovMatrix.
		 */
		// find min off-diagonal (will always be negative, unless all off-diagonals are zero.)
		double min = Double.MAX_VALUE;
		if (rateParameters[startIndex]>=0) {
			for (int row=0; row<dim; row++) {
				for (int col=0; col<dim; col++) {
					if (col != row) min = Math.min(min,rateStore[row][col]);
				}
			}
		} else {
			for (int row=0; row<dim; row++) {
				for (int col=0; col<dim; col++) {
					if (col != row) min = Math.min(min,-rateStore[row][col]);
				}
			}		
		}
		/*
		 *  If we call current matrix X', we want
		 *  X = k X' 
		 *  with k chosen such that minOffDiagonal(X) = -1,
		 *  then final rate matrix is
		 *  Q = (A+rX)/3
		 *  where r = radial parameter (first parameter) and A = Jukes Cantor rate matrix 
		 *  (and always-present basis matrix)
		 *  A = (-3  1  1  1)
		 *      ( 1 -3  1  1)
		 *      ( 1  1 -3  1)
		 *      ( 1  1  1 -3)
		 *     
		 */
		// 'min' is always non-positive, rP[sI] is always non-negative, hence scale is non-negative
		double scale = (min==0) ? 0 : -rateParameters[startIndex]/min/3; // avoid NaN if min==0.
		for (int row=0; row<dim; row++) {
			for (int col=0; col<dim; col++) {				
				if (col == row) {
					rateStore[row][col]=rateStore[row][col]*scale-1;
				} else {
					rateStore[row][col]=rateStore[row][col]*scale+1./3;
				}
			}
		}		
	}
	
	/*
	 * Technically, the allowable range for all parameters is -1 to 1. However, 
	 * r=1 (parameter 0) gives a rate matrix with zeros in it, which may lead to zero likelihood (hence
	 * infinite log likelihood) and ConjugateDirectionSearch in particular is both prone to
	 * evaluate on the boundary, and is unable to handle infinities. So I hack it a bit by
	 * setting the maximum to a little less than one.
	 * 
	 * Also, if r=0, all other parameters are irrelevant (we have the Jukes Cantor rate matrix.)
	 * I give it a non-zero minimum so that the other
	 * parameters will always have some effect, else we might just get stuck on parameter[0]=0
	 * and never be able to move off it as we never try a favourable direction.
	 * 
	 */
	@Override
	public double getRateParameterLowerBound(int parameter) { return (parameter==0) ? 1e-5 : 0; }
	@Override
	public double getRateParameterUpperBound(int parameter) { return (parameter==0) ? 0.99999 : 1; }
	/*
	 *  Choice of default parameters is pretty arbitrary
	 *  0, 0.5 and 1 are poor choices as they lead to zeros and ones in basis weights which may
	 *  lead to a poorly conditioned eigenvalues matrix.
	 */
	@Override
	public void getDefaultRateParameters(double[] parameterStore, int startIndex) {
		int n = getNumberOfRateParameters();
		for (int i=startIndex; i<startIndex+n; i++) parameterStore[i] = 0.3;
	}

	// I've been lazy about superfluous ',' and end of row.
	@Override
	public void printObject(PrintStream out, String prefixString) {
		out.printf("%sLieMarkovRateMatrix %s:\n",prefixString, name_);
		StringBuffer[][] matrix = new StringBuffer[4][4];
		for (int i=0; i<4; i++) for (int j=0; j<4; j++) matrix[i][j] = new StringBuffer();
		for (int i=0; i<=nParam_; i++) {
			matrices_[i].applyToStrings(matrix);
		}
		int maxlen=0;
		for (int i=0; i<4; i++) for (int j=0; j<4; j++) maxlen = Math.max(maxlen, matrix[i][j].length());
		String format = "%-"+Integer.toString(maxlen)+"s, ";
		for (int row=0; row<4; row++) {
			out.print(prefixString+'{');
			for (int col=0; col<4; col++) {
				out.printf(format,matrix[row][col].toString());
			}
			out.printf("}%n");
		}	
	}
}
