package lieAlgebra;

import pal.misc.Parameterized;
/*
 * 
 * Licensed under the Creative Commons Zero (CC0) license (https://creativecommons.org/publicdomain/zero/1.0/)
 * 
 */

/*
 * An n-sphere is the set of points equidistant from a center point in
 * an n+1 dimensional space. (Here, we always take the center point to be the origin.)
 * 
 * This class provides conversion between specifying the point by n angles,
 * and the n+1 Euclidean coordinates of that point.
 * 
 *  I've created the class by copying and editing OctahedralVector, which does a similar job
 *  (but instead for points on the surface of an n-octahedron). This is lazy:
 *  the correct way to encapsulate the similarities it that they should both 
 *  subclass some abstract class.
 *  
 *  Notation:
 *  "parameters" refers to the angles
 *  "coordinates" refers to the Euclidean coordinates.
 *  
 *  To make the parameter ranges tidier, I've rescaled the angles, so all
 *  parameters are in the range [0,1].
 *  
 *  Like OctahedralVector, can be used for its static methods, or can have
 *  an object which provides a 'parameters' and a 'coordinates' view on the
 *  same data point.
 */
public class SphericalVector implements Parameterized {
	double[] coordinates_; // 'y' in the comments above
	double[] parameters_; // 'x' in the comments above
	int n_; // = vector_.length;
	
	/**
	 * @param dimensions: a vector with non-negative elements summing to one
	 * 
	 * NOTE: dimensions array is call-by-reference. The passed array can
	 * have its contents changed by this object. 
	 */
	public SphericalVector(double[] coordinates) {
		allocateSpace(coordinates.length);
		setCoordinates(coordinates);
		coordinatesToParameters(coordinates_, parameters_, 0);
	}
	
	public SphericalVector clone() {
		return new SphericalVector(this.coordinates_);
	}
	
	/**
	 * @return The Spherical vector
	 */
	public double[] getCoordinates() {
		return coordinates_;
	}
	/**
	 * copies the Spherical vector into the store.
	 */
	public void getCoordinates(double[] store) {
		parametersToCoordinates(parameters_,store);
	}
	/**
	 * Set the Spherical vector (validity of vector not checked.)
	 * @param coordinates
	 * 
	 * Note: Unlike the constructor, this copies the input (call by value).
	 * Is there some logical reason for this, or is it an inadvertent inconsistency?
	 */
	public void setCoordinates(double[] coordinates) {
		allocateSpace(coordinates.length);
		coordinates_ = coordinates.clone();
		coordinatesToParameters(coordinates_, parameters_, 0);
	}

	/**
	 * @return The length of the stochastic vector (one more than the number of parameters)
	 */
	// c.f. getNumParameters()
	public int length() {
		return n_;
	}
	/**
	 * If vector length has changed, updates n_ and allocates a new parameters_ array.
	 * Does NOT change coordinates_ array, as whatever is calling allocateSpace takes care of this.
	 * @param newLen
	 */
	private void allocateSpace(int newLen) {
		if (newLen != n_) {
			n_ = newLen;
			parameters_ = new double[n_-1];
		}
	}
	
	/**
	 * Methods for Parameterized interface:
	 */
	public double getDefaultValue(int n) { return 0.5; } // A pretty arbitrary choice. There is no 'neutral' parameter setting.
	public double getLowerLimit(int n)   { return 0; }
	public double getUpperLimit(int n)   { return 1; }
	public int getNumParameters()        { return n_-1; }
	public double getParameter(int n)    { return parameters_[n]; }
	public void setParameter(double param, int n) {
		parameters_[n] = param;
		parametersToCoordinates(parameters_,coordinates_);
	}
	public void setParameterSE(double paramSE, int n) { throw new RuntimeException("Not implemented");	}
	
	
	/* **************************************************************************************************************
	 * The static methods:
	 */
	
	/**
	 * The most general method. 
	 * @param n The length of the Spherical vector to be generated
	 * @param parameters A vector containing within it the n-1 parameters to be mapped to the Spherical vector 
	 * @param parametersOffset The starting point within the parameters vector 
	 * @param coordinates A vector to store the results within
	 * @param coordinatesOffset Where within vector to start storing results.
	 */
	public static void parametersToCoordinates(int n, double[] parameters, int parametersOffset, double[] coordinates, int coordinatesOffset) {
		double residualRadius = 1;
		double value;
		for (int i=0; i<n-2; i++) {
			value = residualRadius * Math.cos(Math.PI*parameters[i+parametersOffset]);
			coordinates[i+coordinatesOffset] = value;
			residualRadius *= Math.sin(Math.PI*parameters[i+parametersOffset]);
		}
		value = residualRadius * Math.cos(2*Math.PI*parameters[n-2+parametersOffset]);
		coordinates[coordinatesOffset+n-2] = value;
		value = residualRadius * Math.sin(2*Math.PI*parameters[n-2+parametersOffset]);
		coordinates[coordinatesOffset+n-1] = value;		
	}
	
	/**
	 * Write Spherical vector into the start of provided storage space,
	 * using all of the provided parameters vector. 
	 * @param parameters
	 * @param coordinates
	 */
	public static void parametersToCoordinates(double[] parameters, double[] coordinates) {
		parametersToCoordinates(parameters.length+1, parameters, 0, coordinates, 0);
	}
	
	public static double[] parametersToCoordinates(double[] parameters) {
		double[] coordinates = new double[parameters.length+1];
		parametersToCoordinates(parameters, coordinates);
		return coordinates;
	}
	
	/**
	 * Inverse mapping: given an Spherical vector, find the corresponding parameters.
	 * @param n The length of the Spherical vector supplied
	 * @param coordinates A vector containing within it the Spherical vector 
	 * @param coordinatesOffset Where to find the Spherical vector within 'coordinates'
	 * @param parameters A vector to store the results within
	 * @param parametersOffset Where within parameters array to start storing results. (Will store (n-1) numbers.)
	 * 
	 * If coordinates to not have squared values summing to one, will still produce a result
	 * (without throwing an exception) which will, on reverse transformation, match the input
	 * for as many initial entries as is possible.
	 */
	public static void coordinatesToParameters(int n, double[] coordinates, int coordinatesOffset, double[] parameters, int parametersOffset) {
		// TODO: implement
		throw new RuntimeException("Unimplemented");
	}
	/**
	 * Inverse mapping: given an Spherical vector, find the corresponding parameters.
	 * @param coordinates The Spherical vector (of length n)
	 * @param parameters A vector to store the parameters within 
	 * @param parametersOffset Where within vector to start storing results. (Will store (n-1) numbers.)
	 */
	public static void coordinatesToParameters(double[] coordinates, double[] parameters, int parametersOffset) {
			coordinatesToParameters(coordinates.length, coordinates, 0, parameters, parametersOffset);
	}
}
